#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from types import ModuleType
from typing import Any, Union, Tuple, Callable, Type
from splunk_add_on_ucc_modinput_test.common import ta_base
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.manager import (
    dependency_manager,
    forge,
    forges,
)
from splunk_add_on_ucc_modinput_test.functional.splunk import (
    SplunkClientBase,
    SplunkConfigurationBase,
)
from splunk_add_on_ucc_modinput_test.functional.vendor import (
    VendorClientBase,
    VendorConfigurationBase,
)

from .constants import BuiltInArg


def bind(
    fn: Callable[..., Any],
    forges: Tuple[Union[forge, forges], ...],
    is_bootstrap: bool,
) -> None:
    for item in forges:
        scope = item.scope
        step_forges = (
            [item] if isinstance(item, forge) else list(item.forge_list)
        )
        dependency_manager.bind(fn, scope, step_forges, is_bootstrap)


def bootstrap(*forges: Union[forge, forges]) -> Callable[..., Any]:
    def bootstrap_dec(fn: Callable[..., Any]) -> Callable[..., Any]:
        bind(fn, forges, is_bootstrap=True)
        return fn

    return bootstrap_dec


def attach(*forges: Union[forge, forges]) -> Callable[..., Any]:
    def attach_dec(fn: Callable[..., Any]) -> Callable[..., Any]:
        bind(fn, forges, is_bootstrap=False)
        return fn

    return attach_dec


def define_vendor_client_argument(
    vendor_client_class: Type[VendorClientBase],
    vendor_class_argument_name: str = BuiltInArg.VENDOR_CLIENT.value,
) -> Callable[..., Any]:
    logger.debug(
        "attach_vendor_client_config vendor_client_class: "
        f"{vendor_client_class} with vendor_class_argument_name "
        + vendor_class_argument_name
    )

    def register_vendor_class_decorator(
        vendor_configuration_class: Type[VendorConfigurationBase],
    ) -> Type[VendorConfigurationBase]:
        dependency_manager.set_vendor_client_class(
            vendor_configuration_class,
            vendor_client_class,
            vendor_class_argument_name,
        )
        return vendor_configuration_class

    return register_vendor_class_decorator


def register_vendor_class(
    vendor_configuration_class: Type[
        VendorConfigurationBase
    ] = VendorConfigurationBase,
) -> Callable[..., Any]:
    logger.debug(f"register_vendor_class {vendor_configuration_class}")

    def register_vendor_class_decorator(
        vendor_client_class: Type[VendorClientBase],
    ) -> Type[VendorClientBase]:
        dependency_manager.set_vendor_client_class(
            vendor_configuration_class,
            vendor_client_class,
            BuiltInArg.VENDOR_CLIENT.value,
        )
        return vendor_client_class

    return register_vendor_class_decorator


def define_splunk_client_argument(
    swagger_client: ModuleType,
    splunk_client_class: Type[SplunkClientBase],
    splunk_class_argument_name: str = BuiltInArg.SPLUNK_CLIENT.value,
) -> Callable[..., Any]:
    logger.debug(
        "attach_splunk_client_config splunk_client_class: "
        f"{splunk_client_class} with "
        f"splunk_class_argument_name {splunk_class_argument_name}"
    )

    def _bind_swagger_client(self: SplunkClientBase) -> None:
        self.ta_service = ta_base.ConfigurationBase(
            swagger_client=swagger_client,
            splunk_configuration=self._splunk_configuration,
        )

    def register_splunk_class_decorator(
        splunk_configuration_class: Type[SplunkConfigurationBase],
    ) -> Type[SplunkConfigurationBase]:
        splunk_client_class._bind_swagger_client = _bind_swagger_client
        dependency_manager.set_splunk_client_class(
            splunk_configuration_class,
            splunk_client_class,
            splunk_class_argument_name,
        )
        return splunk_configuration_class

    return register_splunk_class_decorator


def register_splunk_class(
    swagger_client: ModuleType,
    splunk_configuration_class: Type[
        SplunkConfigurationBase
    ] = SplunkConfigurationBase,
) -> Callable[..., Any]:
    logger.debug(
        f"register_splunk_class swagger_client:{swagger_client} with "
        f"configuration {splunk_configuration_class}"
    )

    def _bind_swagger_client(self: SplunkClientBase) -> None:
        self.ta_service = ta_base.ConfigurationBase(
            swagger_client=swagger_client,
            splunk_configuration=self._splunk_configuration,
        )

    def register_splunk_class_decorator(
        splunk_client_class: Type[SplunkClientBase],
    ) -> Type[SplunkClientBase]:
        splunk_client_class._bind_swagger_client = _bind_swagger_client
        dependency_manager.set_splunk_client_class(
            splunk_configuration_class,
            splunk_client_class,
            BuiltInArg.SPLUNK_CLIENT.value,
        )
        return splunk_client_class

    return register_splunk_class_decorator



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from splunk_add_on_ucc_modinput_test.functional.vendor.configuration import (
    VendorConfigurationBase,
)


class VendorClientBase:
    def __init__(self, vendor_configuration: VendorConfigurationBase) -> None:
        self._vendor_configuration = vendor_configuration

    @property
    def vendor_configuration(self) -> VendorConfigurationBase:
        return self._vendor_configuration

    @property
    def config(
        self,
    ) -> VendorConfigurationBase:  # short alias for vendor_configuration
        return self._vendor_configuration



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Any, Dict, Set
import tempfile
from splunk_add_on_ucc_modinput_test.functional import logger
from typing import IO
import json
from base64 import b64encode
import ssl
import urllib.request
import urllib.parse


class SplunkInstanceFileHelper:
    """Class for Test file helper."""

    class OperationError(Exception):
        pass

    class FileReadError(OperationError):
        pass

    class ExecuteCommandError(OperationError):
        pass

    def __init__(
        self, splunk_url: str, username: str, password: str, base_dir: str = ""
    ) -> None:
        self.file_set: Set[IO[str]] = set()
        self.splunk_url = splunk_url
        self.username = username
        self.password = password
        self.base_dir = base_dir

    def perform_api_operations(
        self, operation: str, payload: Dict[str, str] = {}
    ) -> Any:  # pylint: disable=dangerous-default-value
        """Performs API operations."""

        endpoint = (
            "/servicesNS/nobody/Splunk_TA_Modinput_Test/"
            "Splunk_TA_Modinput_Test_perform_crd_operation/<entry>/"
        )
        api_url = self.splunk_url + endpoint + operation

        payload.update({"output_mode": "json"})
        data = urllib.parse.urlencode(payload).encode("utf-8")

        # Create a request object
        request = urllib.request.Request(api_url, data=data, method="POST")

        # Add basic authentication header
        credentials = f"{self.username}:{self.password}"
        encoded_credentials = b64encode(credentials.encode("utf-8")).decode(
            "utf-8"
        )
        request.add_header("Authorization", f"Basic {encoded_credentials}")

        # Disable SSL verification
        context = ssl._create_unverified_context()

        try:
            with urllib.request.urlopen(request, context=context) as response:
                if response.status > 299:
                    raise SplunkInstanceFileHelper.OperationError(
                        f"Operation: {operation}, failed. \
                            Response Code: {response.status}"
                    )
                return json.loads(response.read().decode("utf-8"))
        except urllib.error.HTTPError as e:
            raise SplunkInstanceFileHelper.OperationError(
                f"Operation: {operation}, failed. \
                    Response Code: {e.code}, Message: {e.reason}"
            )

    def _make_path(self, filepath: str) -> str:
        """Makes path."""
        if self.base_dir == "":
            return filepath
        return f"{self.base_dir}/{filepath}"

    def isfile(self, filepath: str) -> bool:
        """Checks if file."""
        full_path = self._make_path(filepath)
        payload = {"file_path": full_path}
        json_response = self.perform_api_operations("is_file", payload)
        logger.debug(
            f"Checking file object is a file {full_path}: {json_response}"
        )
        entry = json_response["entry"][0]
        if entry["name"] == "success_message":
            return True
        return False

    def isdir(self, dirpath: str) -> bool:
        """Checks if directory."""
        full_path = self._make_path(dirpath)
        payload = {"dir_path": full_path}
        json_response = self.perform_api_operations("is_dir", payload)
        logger.debug(
            f"Checking file object is a folder {full_path}: {json_response}"
        )
        entry = json_response["entry"][0]
        if entry["name"] == "success_message":
            return True
        return False

    def create_file(self, filepath: str, content: str = "") -> None:
        """Creates file."""
        payload = {"file_path": self._make_path(filepath), "data": content}
        self.perform_api_operations("create", payload)

    def overwrite_file(self, filepath: str, content: str = "") -> None:
        """Overwrites file."""
        self.create_file(filepath, content)

    def read_file(self, filepath: str) -> str:
        """Reads file."""
        full_path = self._make_path(filepath)
        payload = {"file_path": full_path}
        json_response = self.perform_api_operations("read", payload)
        logger.debug(f"Read file {full_path} status: {json_response}")
        entry = json_response["entry"][0]
        if entry["name"] != "error_message":
            return entry["content"]["file_content"]

        logger.error(f"Error reading file {full_path}: {json_response}")
        raise SplunkInstanceFileHelper.FileReadError(
            entry["content"]["read_error_message"]
        )

    def append_file(self, filepath: str, content: str = "") -> None:
        """Updates file."""
        payload = {"file_path": self._make_path(filepath), "data": content}
        self.perform_api_operations("update", payload)

    def delete_file(self, filepath: str) -> bool:
        """Deletes file."""
        full_path = self._make_path(filepath)
        payload = {"file_path": full_path}
        json_response = self.perform_api_operations("delete", payload)
        logger.debug(f"Delete file {full_path} status: {json_response}")
        entry = json_response["entry"][0]
        if entry["name"] == "success_message":
            return True
        return False

    def remove_dir(self, dirpath: str) -> bool:
        """Removes directory."""
        full_path = self._make_path(dirpath)
        payload = {"dir_path": full_path}
        json_response = self.perform_api_operations("delete_dir", payload)
        logger.debug(f"Delete folder {full_path} status: {json_response}")
        entry = json_response["entry"][0]
        if entry["name"] == "success_message":
            return True
        return False

    def execute(self, command: str) -> str:
        """Executes shell command."""
        payload = {"command": command}
        json_response = self.perform_api_operations("execute", payload)
        entry = json_response["entry"][0]
        if entry["name"] == "success_message":
            return entry["content"]["output"]

        logger.error(f"Error executiong command {command}: {json_response}")
        raise SplunkInstanceFileHelper.ExecuteCommandError(
            entry["content"]["read_error_message"]
        )

    def clean(self) -> None:
        """Clean method."""
        for temp in self.file_set:
            temp.close()

    def retrieve(self, path: str) -> str:
        """
        if remote, then copy the remote file to local
        if not remot, then return the path itself
        :param path: the file that want to get
        :return: local_path the copied file path
        """
        temp = (
            tempfile.NamedTemporaryFile(  # pylint: disable=consider-using-with
                mode="w+"
            )
        )
        self.file_set.add(temp)
        temp_path = temp.name
        data = self.read_file(path)
        temp.write(data)
        return temp_path



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations
import time
from typing import Callable

from splunk_add_on_ucc_modinput_test.common.splunk_instance import (
    search,
    Configuration,
    SearchState,
    Index,
)
from splunk_add_on_ucc_modinput_test.common.splunk_service_pool import (
    SplunkServicePool,
)
from splunk_add_on_ucc_modinput_test.common.ta_base import ConfigurationBase
from splunk_add_on_ucc_modinput_test.functional.common.splunk_instance_file import (  # noqa: E501
    SplunkInstanceFileHelper,
)
from splunk_add_on_ucc_modinput_test.functional.common.splunk_instance_kvstore import (  # noqa: E501
    SplunkInstanceKVStoreAPI,
)
from splunk_add_on_ucc_modinput_test.typing import ProbeGenType
import logging

logger = logging.getLogger("ucc-modinput-test")


class SplunkClientBase:
    def __init__(
        self, splunk_configuration: Configuration | None = None
    ) -> None:
        self.ta_service: ConfigurationBase | None = None
        self._splunk_configuration = splunk_configuration or Configuration()
        self._bind_swagger_client()

    _bind_swagger_client: Callable[..., None] = lambda: None
    # this method is replaced in inherited class by the decorator
    # splunk_add_on_ucc_modinput_test.functional.decorators.register_splunk_class
    # pass

    @property
    def splunk_configuration(self) -> Configuration:
        return self._splunk_configuration

    @property
    def config(self) -> Configuration:  # short alias for splunk_configuration
        return self._splunk_configuration

    @property
    def splunk(self) -> SplunkServicePool:
        return self._splunk_configuration.service

    @property
    def ta_api(
        self,
    ) -> swagger_client.api.default_api.DefaultApi:  # type: ignore    # noqa: E501, F821
        assert (
            self.ta_service is not None
        ), "Make sure you have decorated inherited client class \
            with @register_splunk_class"
        return self.ta_service._api_instance

    @property
    def instance_epoch_time(self) -> int:
        state = self.search("| makeresults | eval splunk_epoch_time=_time")
        if isinstance(state.results, list) and len(state.results) > 0:
            first_result = state.results[0]
            if (
                isinstance(first_result, dict)
                and "splunk_epoch_time" in first_result
            ):
                return int(first_result["splunk_epoch_time"])
        raise ValueError(
            "Expected a list of dictionaries with 'splunk_epoch_time' key \
                in search results"
        )

    @property
    def default_index(self) -> str:
        state = self.search("* | head 1 | fields index")
        if isinstance(state.results, list) and len(state.results) > 0:
            first_result = state.results[0]
            if isinstance(first_result, dict) and "index" in first_result:
                return first_result["index"]
        raise ValueError(
            "Expected a list of dictionaries with 'index' key in search \
                results"
        )

    def _make_conf_error(self, prop_name: str) -> str:
        return f"Make sure you have '{prop_name}' attribute in your Splunk \
            configuration class {self.config.__class__.__name__}"

    @property
    def remote_file_helper(self) -> SplunkInstanceFileHelper:
        connect = dict(
            splunk_url=f"https://{self.config.host}:{self.config.port}",
            username=self.config.username,
            password=self.config.password,
        )
        return SplunkInstanceFileHelper(**connect)

    @property
    def instance_file_helper(self) -> SplunkInstanceFileHelper:
        assert (
            hasattr(self.config, "home")
            and self.config.splunk_home  # type: ignore
        ), self._make_conf_error("home")
        connect = dict(
            splunk_url=f"https://{self.config.host}:{self.config.port}",
            username=self.config.username,
            password=self.config.password,
            base_dir=self.config.home,
        )
        return SplunkInstanceFileHelper(**connect)

    @property
    def app_file_helper(self) -> SplunkInstanceFileHelper:
        assert hasattr(self.config, "home") and self._make_conf_error("home")
        assert hasattr(self.config, "app_name") and self._make_conf_error(
            "app_name"
        )
        connect = dict(
            splunk_url=f"https://{self.config.host}:{self.config.port}",
            username=self.config.username,
            password=self.config.password,
            base_dir=f"{self.config.home}/etc/apps/{self.config.app_name}",
        )
        return SplunkInstanceFileHelper(**connect)

    def search(self, searchquery: str) -> SearchState:
        return search(service=self.splunk, searchquery=searchquery)

    @property
    def _is_cloud(self) -> bool:
        return "splunkcloud.com" in self.config.host.lower()

    def create_index(self, index_name: str, datatype: str = "event") -> Index:
        return self.config._create_index(
            index_name,
            self.splunk,
            datatype=datatype,
            is_cloud=self._is_cloud,
            acs_stack=self.config.acs_stack if self._is_cloud else None,
            acs_server=self.config.acs_server if self._is_cloud else None,
            splunk_token=self.config.token if self._is_cloud else None,
        )

    def get_index(self, index_name: str) -> Index | None:
        return self.config._get_index(
            index_name,
            self.splunk,
            is_cloud=self._is_cloud,
            acs_stack=self.config.acs_stack if self._is_cloud else None,
            acs_server=self.config.acs_server if self._is_cloud else None,
            splunk_token=self.config.token if self._is_cloud else None,
        )

    def kvstore_api_helper(
        self,
        collection_name: str,
        record_id: str,
    ) -> SplunkInstanceKVStoreAPI:
        assert hasattr(
            self.config, "app_name"
        ), "app_name attribute is not set in Splunk configuration"
        assert hasattr(
            self.config, "app_user"
        ), "app_user attribute is not set in Splunk configuration"
        connect = dict(
            splunk=self.splunk,
            collection_name=collection_name,
            record_id=record_id,
            app_name=self.config.app_name,
            app_user=self.config.app_user,
        )
        return SplunkInstanceKVStoreAPI(**connect)

    def search_probe(
        self,
        probe_spl: str,
        *,
        verify_fn: Callable[[SearchState], bool] | None = None,
        timeout: int = 300,
        interval: int = 5,
        probe_name: str | None = "probe",
    ) -> ProbeGenType:
        """
        Probe state by search until it returns verify function return true.
        Default verify function checks for non empty search result.
        @param probe_spl: Splunk search query to execute for each check
                iteration.
        @param verify_fn: function to verify search state.
        @param timeout: how long in seconds to wait for positive result.
        @param interval: interval to repeat search.
        @return: True, if probe conditions met before expiration,
            otherwise False.
        """

        def non_empty_result(state: SearchState) -> bool:
            return state.result_count > 0

        if verify_fn is None:
            verify_fn = non_empty_result
        else:
            assert callable(verify_fn)

        if probe_name is None:
            probe_name = verify_fn.__name__

        start_time = time.time()
        logger.debug(probe_spl)
        expire = start_time + timeout
        while time.time() < expire:
            search_state = self.search(searchquery=probe_spl)
            if verify_fn(search_state):
                logger.debug(
                    f"{probe_name} has successfully finished after "
                    f"{time.time() - start_time} seconds"
                )
                return True
            logger.debug(
                f"{probe_name} is negative after "
                f"{time.time()-start_time} seconds"
            )
            yield interval

        logger.info(f"{probe_name} is still negative after {timeout} seconds")
        return False

    def repeat_search_until(
        self,
        spl: str,
        *,
        condition_fn: Callable[[SearchState], bool] | None = None,
        timeout: int = 300,
        interval: int = 5,
    ) -> SearchState | None:
        """
        Repeats search until condition function returns True.
        Default condition function checks for non empty search result.
        @param spl: Splunk search query to execute.
        @param condition_fn: function to verify search state to stop search
                iterations.
        @param timeout: how long in seconds to wait for positive result of
                condition function.
        @param interval: interval to repeat search if condition function is
                negative.
        @return: SearchState object.
        """

        it = self.search_probe(
            probe_spl=spl,
            verify_fn=condition_fn,
            timeout=timeout,
            interval=interval,
            probe_name=None,
        )

        try:
            while True:
                wait = next(it)
                time.sleep(wait)
        except StopIteration as si:
            return si.value



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

fh = logging.FileHandler("ucc_modinput_test.log")
fh.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s - %(process)d - %(filename)s - \
        %(funcName)s - %(message)s"
)
fh.setFormatter(formatter)
logger = logging.getLogger("ucc-modinput-test")
logger.addHandler(fh)
logging.root.propagate = False
logger.setLevel(logging.DEBUG)

logger.debug("Logger set")



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Generator

if TYPE_CHECKING:
    from splunk_add_on_ucc_modinput_test.typing import (
        ForgeFnType,
        ExecutableKeyType,
        ArtifactsType,
    )

import threading
import inspect
import contextlib
import time
from dataclasses import dataclass, replace
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.entities.executable import (
    ExecutableBase,
)


@dataclass
class ForgeExecData:
    id: str
    teardown: Generator[Any, None, None] | None
    kwargs: dict[str, Any]
    result: ArtifactsType
    errors: list[str]
    count: int
    lock: threading.Lock
    is_teardown_executed: bool = False

    def __init__(
        self,
        id: str,
        teardown: Generator[Any, None, None] | None,
        kwargs: dict[str, Any],
        result: ArtifactsType,
        errors: list[str],
        count: int,
    ) -> None:
        self.lock = threading.Lock()
        self.id = id
        self.teardown = teardown
        self.kwargs = kwargs
        self.errors = errors
        self.count = count

        try:
            # make a copy of forge execution data if possible
            self.result = deepcopy(result)
        except TypeError:
            # copy is not possible, saving forge execution data by reference
            self.result = result

            summary = self.summary("\t")
            logger.warning(
                f"deepcopy of forge execution data is not possible, saving is by reference.{summary}"
            )

    def summary(self, offset: str = "") -> str:
        s = f"\n{offset}Teardown summary:"
        s += f"\n{offset}\tdata.id: {self.id},"
        s += f"\n{offset}\tdata.count={self.count},"
        s += f"\n{offset}\tdata.is_teardown_executed={self.is_teardown_executed}"
        s += f"\n{offset}\tdata.teardown={self.teardown}"
        s += f"\n{offset}\tdata.kwargs={self.kwargs}"
        s += f"\n{offset}\tdata.result={self.result}"
        return s


class ForgePostExec:
    def __init__(self) -> None:
        self.lock = threading.Lock()
        self._exec_store: dict[str, ForgeExecData] = {}
        self._teardown_is_blocked = False

    def summary(self, data: ForgeExecData) -> str:
        s = data.summary()
        s += f"\n\tteardown_is_blocked={self._teardown_is_blocked}"
        return s

    def block_teardown(self) -> None:
        self._teardown_is_blocked = True

    def unblock_teardown(self) -> None:
        self._teardown_is_blocked = False
        self.exec_ready_teardowns()

    def exec_ready_teardowns(self) -> None:
        for data in self._exec_store.values():
            self.exec_teardown_if_ready(data)

    def exec_teardown_if_ready(self, data: ForgeExecData) -> bool:
        with data.lock:
            can_execute = (
                data.count == 0
                and not self._teardown_is_blocked
                and not data.is_teardown_executed
            )
            logger.debug(
                f"CAN EXECUTE TEARDOWN {can_execute}:{self.summary(data)}"
            )
            if can_execute:
                self.execute_teardown(data)
        return can_execute

    def add(
        self,
        id: str,
        teardown: Generator[Any, None, None] | None,
        kwargs: dict[str, Any],
        result: ArtifactsType,
        errors: list[str],
    ) -> None:
        if id not in self._exec_store:
            with self.lock:
                data = ForgeExecData(id, teardown, kwargs, result, errors, 1)
                self._exec_store[id] = data
            logger.debug(f"REGISTER TEARDOWN {id}: {self.summary(data)}")
        else:
            self.reuse(id)

    def get(self, id: str) -> ForgeExecData:
        data = self._exec_store.get(id)
        assert data
        return replace(data)

    def reuse(self, id: str) -> None:
        data = self._exec_store.get(id)
        assert data
        logger.debug(f"REUSE TEARDOWN {id}:{self.summary(data)}")
        with data.lock:
            data.count += 1

    def get_teardown(self, id: str) -> Generator[Any, None, None] | None:
        data = self._exec_store.get(id)
        assert data
        return data.teardown

    def get_result(self, id: str) -> object:
        data = self._exec_store.get(id)
        assert data
        return data.result

    def get_count(self, id: str) -> int:
        data = self._exec_store.get(id)
        assert data
        return data.count

    def is_teardown_executed(self, id: str) -> bool:
        data = self._exec_store.get(id)
        assert data
        return data.is_teardown_executed

    def list(self) -> tuple[ForgeExecData, ...]:
        return tuple(self._exec_store.values())

    def execute_teardown(self, data: ForgeExecData) -> None:
        teardown = data.teardown
        if inspect.isgenerator(teardown):
            with contextlib.suppress(StopIteration):
                next(teardown)
        elif callable(teardown):
            teardown()
        else:
            pass
        data.is_teardown_executed = True

    def dereference_teardown(self, id: str) -> bool:
        data = self._exec_store.get(id)
        if data is None:
            return False

        logger.debug(f"BEFORE EXECUTE TEARDOWN {id}:{self.summary(data)}")

        with data.lock:
            data.count -= 1

        teardown_start_time = time.time()
        executed = self.exec_teardown_if_ready(data)
        if executed:
            logger.info(
                f"Teardown has been executed successfully, time taken: {time.time() - teardown_start_time} seconds:{self.summary(data)}"
            )
        logger.debug(f"AFTER EXECUTE TEARDOWN {id}:{self.summary(data)}")
        return executed


class FrameworkForge(ExecutableBase):
    def __init__(
        self,
        function: ForgeFnType,
        scope: str,
    ) -> None:
        super().__init__(function)
        self._function: ForgeFnType = function
        self._scope = scope
        self.tests: set[ExecutableKeyType] = set()
        self._executions = ForgePostExec()

    @property
    def key(self) -> ExecutableKeyType:
        key_value = list(super().key)
        key_value.append(self._scope)
        return tuple(key_value)

    def set_scope(self, scope: str) -> None:
        self._scope = scope

    @property
    def scope(self) -> str:
        return self._scope

    @property
    def tests_keys(self) -> list[ExecutableKeyType]:
        return list(self.tests)

    @property
    def name(self) -> str:
        return self.key[1]

    @property
    def path(self) -> str:
        return self.source_file

    @property
    def full_path(self) -> str:
        return "::".join(self.key[:2])

    @property
    def executions(self) -> tuple[ForgeExecData, ...]:
        return self._executions.list()

    def block_teardown(self) -> None:
        self._executions.block_teardown()

    def unblock_teardown(self) -> None:
        self._executions.unblock_teardown()

    def teardown(self, id: str) -> bool:
        return self._executions.dereference_teardown(id)

    def register_execution(
        self,
        id: str,
        *,
        teardown: Generator[Any, None, None] | None,
        kwargs: dict[str, Any],
        result: ArtifactsType,
        errors: list[str],
    ) -> None:
        self._executions.add(id, teardown, kwargs, result, errors)

    def reuse_execution(self, prev_exec_id: str) -> None:
        self._executions.reuse(prev_exec_id)

    def __contains__(self, test_key: ExecutableKeyType) -> bool:
        return test_key in self.tests

    def unlink_test(self, test_key: ExecutableKeyType) -> None:
        self.tests.discard(test_key)

    def link_test(self, test_key: ExecutableKeyType) -> None:
        if test_key not in self.tests:
            self.tests.add(test_key)

    def __repr__(self) -> str:
        s = repr(self._function)
        return s.replace("<function", "<dependency function")
